import numpy as np
from pylj import util, force, sample


def initialise(number_of_particles, arrangement, threshold, max_steps):
    """Initial particle positions.

    Parameters
    ----------
    number_of_particles: int
        Number of particles in the system.
    arrangement: func
        A function to set the arrangement of the particles.

    Returns
    -------
    Particle array
        An array containing the initial particle positions, velocities and accelerations set.
    System
        System information.
    """
    system = util.System(number_of_particles, 0, 16., 0, threshold=threshold, max_steps=max_steps)
    particles = arrangement(system)
    return particles, system


def steepest_descent(particles, system, sample_system, alpha):
    previous_energy = system.threshold + 1
    current_energy = 0
    counter = 0
    while counter < system.max_steps:
        if np.abs(previous_energy - current_energy) < system.threshold:
            break
        particles = force.calculate_energy_and_force(particles, system)
        energy = []
        for i in range(0, len(particles)):
            energy.append(particles[i].energy)
        k = np.argmax(energy)
        xforce = particles[k].xforce
        yforce = particles[k].yforce
        previous_energy = current_energy
        current_energy = np.sum(energy)
        if np.isnan(current_energy) and sample_system:
            sample_system.update(particles, system, '\nNAN ERROR'.format(counter))
            break
        if counter % 10 == 0 and sample_system:
            sample_system.update(particles, system, 'EM - Steps = {:d}\nEnergy change = {:.3e}'.format(counter, np.abs(
                previous_energy - current_energy)))
        particles[k].xpos = particles[k].xpos - (xforce * alpha)
        particles[k].xpos = particles[k].xpos % system.box_length
        particles[k].ypos = particles[k].ypos - (yforce * alpha)
        particles[k].ypos = particles[k].ypos % system.box_length
        counter += 1
    if sample_system and not np.isnan(current_energy):
        sample_system.update(particles, system, 'EM - Steps = {:d}\nEnergy change = {:.3e}'.format(counter,np.abs(
            previous_energy - current_energy)))
    return particles



import numpy as np
import matplotlib.pyplot as plt
import time

"""
These are a series of classes related to how the system should be sampled.
"""

class Scattering(object):
    """The scattering class will plot the particle positions, radial distribution function, instantaneous pressure
    and the scattering profile (determined as a fft for the rdf).

    Parameters
    ----------
    system: System
        The whole system information. 
    """
    def __init__(self, system):
        fig, ax = environment(4) 
        self.average_rdf = []
        self.r = []
        self.average_diff = []
        self.q = []

        setup_cellview(ax[0, 0], system)
        setup_rdfview(ax[0, 1], system)
        setup_diffview(ax[1, 1])
        setup_pressureview(ax[1, 0])

        ax[1, 0].plot([0], color='#34a5daff')

        plt.tight_layout()
        self.ax = ax
        self.fig = fig

    def update(self, system):
        update_cellview(self.ax[0, 0], system)
        update_rdfview(self.ax[0, 1], system, self.average_rdf, self.r)
        update_diffview(self.ax[1, 1], system, self.average_diff, self.q)
        update_pressureview(self.ax[1, 0], system)
        self.fig.canvas.draw()

    def average(self):
        gr = np.average(self.average_rdf, axis=0)
        x = np.average(self.r, axis=0)
        line = self.ax[0, 1].lines[0]
        line.set_xdata(x)
        line.set_ydata(gr)
        self.ax[0, 1].set_ylim([0, np.amax(gr) + 0.5])
        self.fig.canvas.draw()

        #self.step_text.set_text('Average')

        iq = np.average(self.average_diff, axis=0)
        x = np.average(self.q, axis=0)
        line = self.ax[1, 1].lines[0]
        line.set_ydata(iq)
        line.set_xdata(x)
        self.ax[1, 1].set_ylim([np.amin(iq) - np.amax(iq) * 0.05, np.amax(iq) + np.amax(iq) * 0.05])
        self.ax[1, 1].set_xlim([np.amin(x), np.amax(x)])


class JustCell(object):
    def __init__(self, system):
        fig, ax = plt.subplots(figsize=(4.5, 4.5))

        ax.plot([0]*10, 'o', markersize=14, markeredgecolor='black')
        ax.set_xlim([0, system.box_length])
        ax.set_ylim([0, system.box_length])
        ax.set_xticks([])
        ax.set_yticks([])
        self.temp_text = ax.text(0.98, 0.02, '', transform=ax.transAxes, fontsize=16, horizontalalignment='right',
                                 verticalalignment='bottom')

        plt.tight_layout()

        self.ax = ax
        self.fig = fig

    def update(self, particles, system, text):
        x3 = np.array([])
        y3 = np.array([])
        for i in range(0, particles.size):
            x3 = np.append(x3, particles[i].xpos)
            y3 = np.append(y3, particles[i].ypos)

        line2 = self.ax.lines[0]
        line2.set_ydata(y3)
        line2.set_xdata(x3)
        self.temp_text.set_text(text)
        self.fig.canvas.draw()


class EnergyMinimisation(object):
    def __init__(self, system, energy):
        fig, ax = plt.subplots(1, 2, figsize=(9, 4.5))

        ax[1].plot([0] * 20)
        ax[1].set_ylabel('$Energy$', fontsize=16)
        ax[1].set_xlabel('$Step$', fontsize=16)
        self.step_text = ax[1].text(0.98, 0.95, 'Current energy={:.2e}'.format(energy), transform=ax[1].transAxes,
                                    fontsize=12, horizontalalignment='right', verticalalignment='bottom')

        ax[0].plot([0] * 20, 'o', markersize=14, markeredgecolor='black')
        ax[0].set_xlim([0, system.box_length])
        ax[0].set_ylim([0, system.box_length])
        ax[0].set_xticks([])
        ax[0].set_yticks([])

        plt.tight_layout()
        self.ax = ax
        self.fig = fig
        self.energy_array = []

    def update(self, particles, system, energy):
        self.energy_array.append(energy)

        line = self.ax[1].lines[0]
        line.set_xdata(np.arange(0, len(self.energy_array)))
        line.set_ydata(self.energy_array)
        self.ax[1].set_ylim([np.amin(self.energy_array) - 0.1 * np.amax(self.energy_array),
                             np.amax(self.energy_array) + 0.1 * np.amax(self.energy_array)])
        self.ax[1].set_xlim([0, len(self.energy_array)])
        self.step_text.set_text('Current energy={:.2e}'.format(energy))

        x3 = np.array([])
        y3 = np.array([])
        for i in range(0, particles.size):
            x3 = np.append(x3, particles[i].xpos)
            y3 = np.append(y3, particles[i].ypos)

        line2 = self.ax[0].lines[0]
        line2.set_ydata(y3)
        line2.set_xdata(x3)

        self.fig.canvas.draw()



class ShowForce(object):
    def __init__(self, system):
        fig, ax = plt.subplots(figsize=(4.5, 4.5))

        ax.plot([-1]*10, '-', color='k')
        ax.plot([-1]*10, '-', color='k')
        ax.plot([-1]*10, '-', color='k')
        ax.plot([0]*10, 'o', markersize=14, markeredgecolor='black', color='b')
        ax.set_xlim([0, system.box_length])
        ax.set_ylim([0, system.box_length])
        ax.set_xticks([])
        ax.set_yticks([])
        self.text1 = ax.text(0.98, 0.95, 'fx' + '={:.2f}'.format(0), transform=ax.transAxes,
                                 fontsize=12, horizontalalignment='right', verticalalignment='bottom')
        self.text2 = ax.text(0.98, 0.90, 'fy' + '={:.2f}'.format(0), transform=ax.transAxes,
                                 fontsize=12, horizontalalignment='right', verticalalignment='bottom')

        plt.tight_layout()

        self.ax = ax
        self.fig = fig
        self.box = system.box_length

    def update(self, particles, system):
        x3 = np.array([])
        y3 = np.array([])
        for i in range(0, particles.size):
            x3 = np.append(x3, particles[i].xpos)
            y3 = np.append(y3, particles[i].ypos)

        line2 = self.ax.lines[3]
        line2.set_ydata(y3)
        line2.set_xdata(x3)
        self.fig.canvas.draw()

    def draw_force(self, particles, i, j, xf, yf):
        x4 = [particles[i].xpos, particles[j].xpos]
        y4 = [particles[i].ypos, particles[j].ypos]
        dx = particles[i].xpos - particles[j].xpos
        dy = particles[i].ypos - particles[j].ypos
        line3 = self.ax.lines[0]
        line4 = self.ax.lines[1]
        line5 = self.ax.lines[2]
        if np.abs(dx) < self.box / 2.:
            if np.abs(dy) < self.box / 2.:
                line3.set_xdata(x4)
                line4.set_xdata(x4)
                line3.set_ydata(y4)
                y_new2 = [-1, -1]
                line4.set_ydata(y_new2)
            else:
                line3.set_xdata(x4)
                line4.set_xdata(x4)
                if y4[0] > y4[1]:
                    y_new1 = [y4[0], y4[1]+self.box]
                    y_new2 = [y4[0]-self.box, y4[1]]
                else:
                    y_new1 = [y4[0], y4[1] - self.box]
                    y_new2 = [y4[0] + self.box, y4[1]]
                line3.set_ydata(y_new1)
                line4.set_ydata(y_new2)
        else:
            if np.abs(dy) < self.box / 2.:
                if x4[0] > x4[1]:
                    x_new1 = [x4[0], x4[1]+self.box]
                    x_new2 = [x4[0]-self.box, x4[1]]
                else:
                    x_new1 = [x4[0], x4[1] - self.box]
                    x_new2 = [x4[0] + self.box, x4[1]]
                line3.set_xdata(x_new1)
                line4.set_xdata(x_new2)
                line3.set_ydata(y4)
                line4.set_ydata(y4)
            else:
                if x4[0] > x4[1]:
                    x_new1 = [x4[0], x4[1]+self.box]
                    x_new2 = [x4[0]-self.box, x4[1]]
                else:
                    x_new1 = [x4[0], x4[1] - self.box]
                    x_new2 = [x4[0] + self.box, x4[1]]
                line3.set_xdata(x_new1)
                line4.set_xdata(x_new2)
                if y4[0] > y4[1]:
                    y_new1 = [y4[0], y4[1] + self.box]
                    y_new2 = [y4[0] - self.box, y4[1]]
                else:
                    y_new1 = [y4[0], y4[1] - self.box]
                    y_new2 = [y4[0] + self.box, y4[1]]
                line3.set_ydata(y_new1)
                line4.set_ydata(y_new2)
        self.text1.set_text('fx' + '={:.2f}'.format(xf))
        self.text2.set_text('fy' + '={:.2f}'.format(yf))
        self.fig.canvas.draw()
        time.sleep(4)

    def clear_force(self):
        x5 = []
        y5 = []
        line3 = self.ax.lines[0]
        line3.set_xdata(x5)
        line3.set_ydata(y5)
        line4 = self.ax.lines[1]
        line4.set_xdata(x5)
        line4.set_ydata(y5)
        self.text1.set_text('')
        self.text2.set_text('')
        self.fig.canvas.draw()


class RDF(object):
    def __init__(self, system):
        fig, ax = plt.subplots(1, 2, figsize=(9, 4.5))

        ax[1].plot([0] * 20)
        ax[1].set_xlim([0, system.box_length / 2])
        ax[1].set_xticks([])
        ax[1].set_yticks([])
        ax[1].set_ylabel('RDF', fontsize=16)
        ax[1].set_xlabel('r', fontsize=16)
        self.step_text = ax[1].text(0.98, 0.95, 'Time={:.1f}'.format(system.step), transform=ax[1].transAxes,
                                    fontsize=12, horizontalalignment='right', verticalalignment='bottom')

        ax[0].plot([0] * 20, 'o', markersize=14, markeredgecolor='black')
        ax[0].set_xlim([0, system.box_length])
        ax[0].set_ylim([0, system.box_length])
        ax[0].set_xticks([])
        ax[0].set_yticks([])

        plt.tight_layout()
        self.ax = ax
        self.fig = fig
        self.avgr = []
        self.xgr = []

    def update(self, particles, system):
        hist, bin_edges = np.histogram(system.distances, bins=np.arange(0, 12.5, 0.1))
        gr = hist / (system.number_of_particles * (system.number_of_particles / system.box_length ** 2) * np.pi *
                     (bin_edges[:-1] + 0.1 / 2.) * 0.1)
        self.avgr.append(gr)
        x = bin_edges[:-1] + 0.1 / 2
        self.xgr = x

        line = self.ax[1].lines[0]
        line.set_xdata(x)
        line.set_ydata(gr)
        self.ax[1].set_ylim([0, np.amax(gr) + 0.5])
        self.step_text.set_text('Time={:.1f}'.format(system.time))

        x3 = np.array([])
        y3 = np.array([])
        for i in range(0, particles.size):
            x3 = np.append(x3, particles[i].xpos)
            y3 = np.append(y3, particles[i].ypos)

        line2 = self.ax[0].lines[0]
        line2.set_ydata(y3)
        line2.set_xdata(x3)

        self.fig.canvas.draw()

    def average_rdf(self):
        gr = np.average(self.avgr, axis=0)
        x = self.xgr
        line = self.ax[1].lines[0]
        line.set_xdata(x)
        line.set_ydata(gr)
        self.ax[1].set_ylim([0, np.amax(gr) + 0.5])
        self.step_text.set_text('Average')

class Temperature(object):
    def __init__(self, system):
        fig, ax = plt.subplots(1, 2, figsize=(9, 4.5))

        ax[0].plot([0] * 20, 'o', markersize=14, markeredgecolor='black')
        ax[0].set_xlim([0, system.box_length])
        ax[0].set_ylim([0, system.box_length])
        ax[0].set_xticks([])
        ax[0].set_yticks([])
        self.step_text = ax[0].text(0.98, 0.02, '', transform=ax[0].transAxes,
                                       fontsize=12, horizontalalignment='right', verticalalignment='bottom')
        ax[1].plot([0] * 20)
        ax[1].set_ylabel('Temperature', fontsize=16)
        ax[1].set_xlabel('Step', fontsize=16)
        self.temp_text = ax[1].text(0.98, 0.02, 'Temperature={:f}'.format(np.average(system.temp_array)),
                                       transform=ax[1].transAxes, fontsize=12, horizontalalignment='right',
                                       verticalalignment='bottom')

        plt.tight_layout()
        self.ax = ax
        self.fig = fig

    def update(self, particles, system, text):
        self.step_text.set_text(text)

        x3 = np.array([])
        y3 = np.array([])
        for i in range(0, particles.size):
            x3 = np.append(x3, particles[i].xpos)
            y3 = np.append(y3, particles[i].ypos)

        line2 = self.ax[0].lines[0]
        line2.set_ydata(y3)
        line2.set_xdata(x3)

        line1 = self.ax[1].lines[0]
        line1.set_ydata(system.temp_array)
        line1.set_xdata(np.arange(0, len(system.temp_array)))
        self.ax[1].set_xlim(0, len(system.temp_array))
        self.ax[1].set_ylim(np.amin(system.temp_array)-np.amax(system.temp_array) * 0.05,
                               np.amax(system.temp_array)+np.amax(system.temp_array) * 0.05)
        self.temp_text.set_text('Temp={:.3f}+/-{:.3f}'.format(np.average(system.temp_array), np.std(system.temp_array)))

        self.fig.canvas.draw()


class Interactions(object):
    def __init__(self, system):
        fig, ax = environment(4)

        setup_cellview(ax[0, 0], system)
        setup_forceview(ax[0, 1])
        setup_pressureview(ax[1, 0])
        setup_tempview(ax[1, 1])

        plt.tight_layout()
        self.ax = ax
        self.fig = fig

    def update(self, system):
        update_cellview(self.ax[0, 0], system)
        update_forceview(self.ax[0, 1], system)
        update_tempview(self.ax[1, 1], system)
        update_pressureview(self.ax[1, 0], system)

        self.fig.canvas.draw()


def environment(panes):
    if panes == 1:
        fig, ax = plt.subplots(figsize=(4, 4))
    elif panes == 2:
        fig, ax = plt.subplots(1, 2, figsize=(8, 4))
    elif panes == 4:
        fig, ax = plt.subplots(2, 2, figsize=(8, 8))
    return fig, ax
        
def setup_cellview(ax, system):
    xpos = system.particles['xposition']
    ypos = system.particles['yposition']
    ax.plot(xpos, ypos, 'o', markersize=14, markeredgecolor='black', color='#34a5daff')
    ax.set_xlim([0, system.box_length])
    ax.set_ylim([0, system.box_length])
    ax.set_xticks([])
    ax.set_yticks([])

def setup_forceview(ax):
    ax.plot([0], color='#34a5daff')
    ax.set_ylabel('Force', fontsize=16)
    ax.set_xlabel('Time', fontsize=16)

def setup_rdfview(ax, system):
    ax.plot([0], color='#34a5daff')
    ax.set_xlim([0, system.box_length/2])
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_ylabel('RDF', fontsize=16)
    ax.set_xlabel('r', fontsize=16)

def setup_diffview(ax):
    ax.plot([0], color='#34a5daff')
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_ylabel('log(I(q))', fontsize=16)
    ax.set_xlabel('q', fontsize=16)

def setup_pressureview(ax):
    ax.plot([0], color='#34a5daff')
    ax.set_ylabel('Pressure', fontsize=16)
    ax.set_xlabel('Time', fontsize=16)

def setup_tempview(ax):
    ax.plot([0], color='#34a5daff')
    ax.set_ylabel('Temperature', fontsize=16)
    ax.set_xlabel('Time', fontsize=16)

def update_cellview(ax, system):
    x3 = system.particles['xposition']
    y3 = system.particles['yposition']
    line = ax.lines[0]
    line.set_ydata(y3)
    line.set_xdata(x3)

def update_rdfview(ax, system, average_rdf, r):
    hist, bin_edges = np.histogram(system.distances, bins=np.arange(0, 12.5, 0.1))
    gr = hist / (system.number_of_particles * (system.number_of_particles / system.box_length ** 2) * np.pi *
                 (bin_edges[:-1] + 0.1 / 2.) * 0.1)
    average_rdf.append(gr)
    x = bin_edges[:-1] + 0.1 / 2
    r.append(x)

    line = ax.lines[0]
    line.set_xdata(x)
    line.set_ydata(gr)
    ax.set_ylim([0, np.amax(gr) + np.amax(gr) * 0.05])


def update_diffview(ax, system, average_diff, q):
    hist, bin_edges = np.histogram(system.distances, bins=np.arange(0, 12.5, 0.1))
    gr = hist / (system.number_of_particles * (system.number_of_particles / system.box_length ** 2) * np.pi *
                 (bin_edges[:-1] + 0.1 / 2.) * 0.1)
    x2 = np.log10(np.fft.rfftfreq(len(gr))[5:])
    y2 = np.log10(np.fft.rfft(gr)[5:])
    average_diff.append(y2)
    q.append(x2)
    line1 = ax.lines[0]
    line1.set_xdata(x2)
    line1.set_ydata(y2)
    ax.set_ylim([np.amin(y2) - np.amax(y2) * 0.05, np.amax(y2) + np.amax(y2) * 0.05])
    ax.set_xlim([np.amin(x2), np.amax(x2)])

def update_forceview(ax, system):
    line = ax.lines[0]
    line.set_ydata(system.force)
    line.set_xdata(np.arange(0, system.step) * system.timestep_length)
    ax.set_xlim(0, system.step * system.timestep_length) 
    ax.set_ylim(np.amin(system.force)-np.amax(system.force) * 0.05,
                np.amax(system.force)+np.amax(system.force) * 0.05)

def update_tempview(ax, system):
    line = ax.lines[0]
    line.set_ydata(system.temperature)
    line.set_xdata(np.arange(0, system.step) * system.timestep_length)
    ax.set_xlim(0, system.step * system.timestep_length) 
    ax.set_ylim(np.amin(system.temperature)-np.amax(system.temperature) * 0.05,
                     np.amax(system.temperature)+np.amax(system.temperature) * 0.05)

def update_pressureview(ax, system):
    line = ax.lines[0]
    line.set_ydata(system.pressure)
    line.set_xdata(np.arange(0, system.step) * system.timestep_length)
    ax.set_xlim(0, system.step * system.timestep_length)
    ax.set_ylim(np.amin(system.pressure) - np.amax(system.pressure) * 0.05,
                           np.amax(system.pressure) + np.amax(system.pressure) * 0.05)
    



from __future__ import division, absolute_import
import numpy as np

cimport numpy as np

cdef extern from "force.h":
    double compute_pressure(int number_of_particles, const double *xvel, const double *yvel, const double *forces, 
                          double box_length)
    void compute_accelerations(int len_particles, const double *xpos, const double *ypos, double *xacc, double *yacc,
                               double *distances, double *xforce, double *yforce, double box_length, double *force_arr)
    void compute_sd(int len_particles, const double *xpos, const double *ypos, double *energy, double *xforce,
                    double *yforce, double box_length)
    void compute_energy_and_force(int len_particles, const double *xpos, const double *ypos, double *energy,
                                  double *xforce, double *yforce, double *xforcedash, double *yforcedash,
                                  double box_length)
    void compute_force(int len_particles, const double *xpos, const double *ypos, double *xforce, double *yforce,
                       double box_length)
    void scale_velocities(int len_particles, double *xvel, double *yvel, double average_temp, double tempature)


DTYPE = np.float64
ctypedef np.float64_t DTYPE_t

def calculate_pressure(parts, partic, forc, box_l):
    cdef int number_of_particles = parts
    cdef np.ndarray[DTYPE_t, ndim=1] xvel = np.zeros(number_of_particles)
    cdef np.ndarray[DTYPE_t, ndim=1] yvel = np.zeros(number_of_particles)
    cdef np.ndarray[DTYPE_t, ndim=1] forces = np.zeros(len(forc))
    cdef double box_length = box_l
    cdef double pres = 0.

    for i in range(0, number_of_particles):
        xvel[i] = partic['xvelocity'][i]
        yvel[i] = partic['yvelocity'][i]

    pres = compute_pressure(number_of_particles, <const double*>xvel.data, <const double*>yvel.data, <const double*>forces.data, box_length)

    return pres
    

def compute_forces(system):
    cdef int len_particles = system.number_of_particles 
    cdef double box_length = system.box_length
    cdef np.ndarray[DTYPE_t, ndim=1] xpos = np.zeros(len_particles) 
    cdef np.ndarray[DTYPE_t, ndim=1] ypos = np.zeros(len_particles)
    cdef np.ndarray[DTYPE_t, ndim=1] xacc = np.zeros(len_particles)
    cdef np.ndarray[DTYPE_t, ndim=1] yacc = np.zeros(len_particles)
    cdef np.ndarray[DTYPE_t, ndim=1] distances = np.zeros(len(system.distances))
    cdef np.ndarray[DTYPE_t, ndim=1] force_arr = np.zeros(len(system.distances))
    cdef np.ndarray[DTYPE_t, ndim=1] xforce = np.zeros(len_particles)
    cdef np.ndarray[DTYPE_t, ndim=1] yforce = np.zeros(len_particles)

    for i in range(0, len_particles):
        xpos[i] = system.particles['xposition'][i]
        ypos[i] = system.particles['yposition'][i]
        xacc[i] = system.particles['xacceleration'][i]
        yacc[i] = system.particles['yacceleration'][i]
    
    compute_accelerations(len_particles, <const double*>xpos.data, <const double*>ypos.data, <double*>xacc.data,
                          <double*>yacc.data, <double*>distances.data, <double*>xforce.data, <double*>yforce.data,
                          box_length, <double*>force_arr.data)

    for i in range(0, len_particles):
        system.particles['xposition'][i] = xpos[i]
        system.particles['yposition'][i] = ypos[i]
        system.particles['xacceleration'][i] = xacc[i]
        system.particles['yacceleration'][i] = yacc[i]
        system.particles['xforce'][i] = xforce[i]
        system.particles['yforce'][i] = yforce[i]
    
    system.distances = distances
    system.forces = force_arr

    return system

def calculate_sd(particles, system):
    cdef int len_particles = particles.size
    cdef np.ndarray[DTYPE_t, ndim=1] xpos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] ypos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] energy = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] xforce = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] yforce = np.zeros(particles.size)
    cdef double box_length = system.box_length

    for i in range(0, particles.size):
        xpos[i] = particles[i].xpos
        ypos[i] = particles[i].ypos

    compute_sd(len_particles, <const double*> xpos.data, <const double*> ypos.data, <double*> energy.data,
                             <double*>xforce.data, <double*> yforce.data, box_length)

    for i in range(0, particles.size):
        particles[i].energy = energy[i]
        particles[i].xforce = xforce[i]
        particles[i].yforce = yforce[i]

    return particles

def calculate_energy_and_force(particles, system):
    cdef int len_particles = particles.size
    cdef np.ndarray[DTYPE_t, ndim=1] xpos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] ypos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] energy = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] xforce = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] yforce = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] xforcedash = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] yforcedash = np.zeros(particles.size)
    cdef double box_length = system.box_length

    for i in range(0, particles.size):
        xpos[i] = particles[i].xpos
        ypos[i] = particles[i].ypos


    compute_energy_and_force(len_particles, <const double*> xpos.data, <const double*> ypos.data, <double*> energy.data,
                             <double*>xforce.data, <double*> yforce.data, <double*> xforcedash.data,
                             <double*> yforcedash.data, box_length)

    for i in range(0, particles.size):
        particles[i].energy = energy[i]
        particles[i].xforce = xforce[i]
        particles[i].yforce = yforce[i]
        particles[i].xforcedash = xforcedash[i]
        particles[i].yforcedash = yforcedash[i]

    return particles

def get_forces(particles, system):
    cdef int len_particles = particles.size
    cdef np.ndarray[DTYPE_t, ndim=1] xpos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] ypos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] xforce = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] yforce = np.zeros(particles.size)
    cdef double box_length = system.box_length

    for i in range(0, particles.size):
        xpos[i] = particles[i].xpos
        ypos[i] = particles[i].ypos


    compute_force(len_particles, <const double*> xpos.data, <const double*> ypos.data, <double*>xforce.data,
                  <double*> yforce.data, box_length)

    return xforce, yforce


def heat_bath(system, bath_temp):
    cdef int len_particles = system.number_of_particles
    cdef np.ndarray[DTYPE_t, ndim=1] xvel = np.zeros(system.number_of_particles)
    cdef np.ndarray[DTYPE_t, ndim=1] yvel = np.zeros(system.number_of_particles)
    cdef double average_temp = np.average(system.temperature)
    cdef double temperature = bath_temp

    for i in range(0, system.number_of_particles):
        xvel[i] = system.particles['xvelocity'][i]
        yvel[i] = system.particles['yvelocity'][i]

    scale_velocities(len_particles, <double*>xvel.data, <double*>yvel.data, average_temp, temperature)

    for i in range(0, system.number_of_particles):
        system.particles['xvelocity'][i] = xvel[i]
        system.particles['yvelocity'][i] = yvel[i]

    return system



from pylj import md, force, sample, util


def periodic_boundary(number_of_steps, temperature):
    """This is an examplary piece of code to show a single particle travelling across the periodic boundary.

    Parameters
    ----------
    number_of_steps: int
        Number of steps to be taken in the md simulation.
    temperature: float
        Temperature of the simulation
    """
    number_of_particles = 1
    sample_freq = 10
    particles, system = md.initialise(number_of_particles, temperature, 0.01, util.set_particles_random)
    sample_system = sample.JustCell(system)
    for i in range(0, number_of_steps):
        particles, system = force.compute_forces(particles, system)
        particles, system = md.velocity_verlet(particles, system)
        system.time += system.timestep_length
        if system.step % sample_freq == 0:
            sample_system.update(particles, system, '')


import numpy as np
from pylj import force, sample, util


def initialise(number_of_particles, temperature, timestep_length, init_conf):
    """Initial particle positions (simple square arrangment), velocities and get initial forces/accelerations.

    Parameters
    ----------
    number_of_particles: int
        Number of particles in the system.
    temperature: float
        Temperature of the system.
    init_conf: string, optional
        Selection for the way the particles are initially populated. Should be one of
        - 'square'
        - 'random'

    Returns
    -------
    System
        System information.
    """
    system = util.System(number_of_particles, temperature, 16, timestep_length, init_conf=init_conf)
    v = np.sqrt(2 * system.init_temp)
    theta = 2 * np.pi * np.random.randn(system.particles.size)
    system.particles['xvelocity'] = v * np.cos(theta)
    system.particles['yvelocity'] = v * np.sin(theta)
    system = force.compute_forces(system)
    system.temp_sum = 0.
    return system

def reset_histogram(system):
    """Reset the velocity histogram.

    Parameters
    ----------
    system: System
        System parameters.

    Returns
    -------
    System
        System parameters with the velocity bins and temperature summation set to zero.
    """
    system.velocity_bins[:]= 0
    system.temp_sum = 0

def velocity_verlet(system):
    """Update the positions, velocities, get temperature and pressure.

    Parameters
    ----------
    particles: Particle array
        All particles in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle array:
        Particles with updated positions and velocities.
    System:
        Whole system information with new temperature and pressure.
    """
    xposition_store = system.particles['xposition']
    yposition_store = system.particles['yposition']
    system.particles['xposition'], system.particles['yposition'] = update_pos_vv([system.particles['xposition'], system.particles['yposition']], [system.particles['xvelocity'], system.particles['yvelocity']], [system.particles['xacceleration'], system.particles['yacceleration']], system.timestep_length, system.box_length)
    system.particles['xvelocity'], system.particles['yvelocity'] = update_velocities_vv([system.particles['xvelocity'], system.particles['yvelocity']], [system.particles['xacceleration'], system.particles['yacceleration']], system.timestep_length)
    system.particles['xprevious_position'] = xposition_store
    system.particles['yprevious_position'] = yposition_store

    temp = util.calculate_temperature(system.number_of_particles, system.particles)
    pres = force.calculate_pressure(system.number_of_particles, system.particles, system.forces, system.box_length)
    system.temperature = np.append(system.temperature, temp)
    system.pressure = np.append(system.pressure, pres)
    system.force = np.append(system.force, np.sum(system.forces))
    return system

def clear_accelerations(particles):
    """Reset all particle accelerations to 0.

    Parameters
    ----------
    particles: Particle array
        The particles in the system.

    Returns
    -------
    Particle array
        The particles in the system now with zero acceleration.
    """
    particles['xacceleration'] = np.zeros(particles.size)
    particles['yacceleration'] = np.zeros(particles.size)
    return particles

def initialise_from_em(particles, system, temperature, timestep_length):
    system.temperature = temperature
    system.timestep_length = timestep_length
    v = np.sqrt(2 * system.temperature)
    for i in range(0, system.number_of_particles):
        theta = 2 * np.pi * np.random.randn()
        particles[i].xvel = v * np.cos(theta)
        particles[i].yvel = v * np.sin(theta)
    particles, system = force.compute_forces(particles, system)
    system = reset_histogram(system)
    return particles, system


def update_pos_verlet(particle, system):
    """Update the positions of a given particle based on the integrator.

    Parameters
    ----------
    particle: Particle
        A particle in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle
        Particle with updated positions.
    """
    particle.xpos = particle.xpos - particle.xpos_prev + particle.xacc * system.timestep_length * system.timestep_length
    particle.ypos = particle.ypos - particle.ypos_prev + particle.yacc * system.timestep_length * system.timestep_length
    particle.xpos = particle.xpos % system.box_length
    particle.ypos = particle.ypos % system.box_length
    return particle


def update_pos_vv(positions, velocities, accelerations, timestep_length, box_length):
    """Update the positions of a given particle based on the integrator.

    Parameters
    ----------
    particle: Particle
        A particle in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle
        Particle with updated positions.
    """
    positions[0] += velocities[0] * timestep_length + 0.5 * accelerations[0] * timestep_length * timestep_length
    positions[1] += velocities[1] * timestep_length + 0.5 * accelerations[1] * timestep_length * timestep_length
    positions[0] = positions[0] % box_length
    positions[1] = positions[1] % box_length
    return positions[0], positions[1] 


def update_velocities_verlet(particle, system):
    """Update the velocities of a given particles based on the acceleration.

    Parameters
    ----------
    particle: Particle
        A particle in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle
        Particle with updated velocities.
    """
    particle.xvel = (particle.xpos - particle.xpos_prev) / (2 * system.timestep_length)
    particle.yvel = (particle.ypos - particle.ypos_prev) / (2 * system.timestep_length)
    return particle


def update_velocities_vv(velocities, accelerations, timestep_length):
    """Update the velocities of a given particles based on the acceleration.

    Parameters
    ----------
    particle: Particle
        A particle in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle
        Particle with updated velocities.
    """
    velocities[0] += 0.5 * accelerations[0] * timestep_length
    velocities[1] += 0.5 * accelerations[1] * timestep_length
    return velocities[0], velocities[1] 

def verlet(particles, system):
    """Update the positions, velocities, get temperature and pressure.

    Parameters
    ----------
    particles: Particle array
        All particles in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle array:
        Particles with updated positions and velocities.
    System:
        Whole system information with new temperature and pressure.
    """
    system = reset_histogram(system)
    system.step += 1
    for i in range(0, system.number_of_particles):
        particles[i] = update_pos_verlet(particles[i], system)
        particles[i] = update_velocities_verlet(particles[i], system)
    particles, system = util.calculate_temperature(particles, system)
    system = util.calculate_pressure(system)
    return particles, system

def update_velocity_bins(particle, velocity_bins, max_vel):
    """Updates the velocity bins. It is not clear if this is completely required if the velocity bins are not
    being plotted.
    
    Parameters
    ----------
    particle: Particle
        A particle in the system.
    system: System
        Whole system information.
    
    Returns
    -------
    System
        Whole system information with the velocity bin for particle updated.
    float
        The total velocity for a given particle.
    """
    v = np.sqrt(particle['xvelocity'] * particle['xvelocity'] + particle['yvelocity'] * particle['yvelocity'])
    bin_s = int(len(velocity_bins) * v / max_vel)
    if bin_s < 0:
        bin_s = 0
    if bin_s >= len(velocity_bins):
        bin_s = len(velocity_bins) - 1
    velocity_bins[bin_s] += 1
    return velocity_bins, v




import numpy as np
from pylj import md, force


class System:
    """Whole system.

    This class stores a large amount of information about the job that is being run.

    Parameters
    ----------
    number_of_particles: int
        Number of particles to be simulated.
    temperature: float
        The initial temperature of the system.
    box_length: float
        Size of the simulation cell.
    timestep_length: float
        Duration of timestep in MD integration.
    max_vel: float, optional
        Maximum velocity allowed in the velocity rebinning stage.
    init_conf: string, optional
        Selection for the way the particles are initially populated. Should be one of
        - 'square'
        - 'random'
    """
    def __init__(self, number_of_particles, temperature, box_length, timestep_length, 
                 max_vel=4, init_conf='square'):
        self.number_of_particles = number_of_particles
        self.init_temp = temperature
        self.box_length = box_length
        self.timestep_length = timestep_length
        self.max_vel = max_vel
        if init_conf == 'square':
            self.square()
        elif init_conf == 'random':
            self.random()
        else:
            raise NotImplementedError('The initial configuration type {} is not recognised. Available options are: square or random'.format(init_conf))
        self.step = 0
        self.time = 0.
        self.temp_sum = 0.
        self.distances = np.zeros(self.number_of_pairs())
        self.forces = np.zeros(self.number_of_pairs())
        self.velocity_bins = np.zeros(500)
        self.temperature = np.array([])
        self.pressure = np.array([])
        self.force = np.array([])

    def number_of_pairs(self):
        return int((self.number_of_particles - 1) * self.number_of_particles / 2)
        
    def square(self):
        """Set the initial particle positions on a square lattice.

        Returns
        -------
        Particle array
            The particles with positions on a square lattice.
        """
        part_dt = particle_dt()
        self.particles = np.zeros(self.number_of_particles, dtype=part_dt)
        m = int(np.ceil(np.sqrt(self.number_of_particles)))
        d = self.box_length / m
        n = 0
        for i in range(0, m):
            for j in range(0, m):
                if n < self.number_of_particles:
                    self.particles[n]['xposition'] = (i + 0.5) * d
                    self.particles[n]['yposition'] = (j + 0.5) * d
                    n += 1
    
    def random(self):
        """Set the initial particle positions in a random arrangement.
        
        Returns
        -------
        Particle array
            The particles with position set randomly.
        """
        part_dt = particle_dt()
        self.particles = np.zeros(self.number_of_particles, dtype=part_dt)
        self.particles['xposition'] = np.random.uniform(0, self.box_length, self.number_of_particles)
        self.particles['yposition'] = np.random.uniform(0, self.box_length, self.number_of_particles)   

def pbc_correction(d, l):
    """Test and correct for the periodic boundary condition.

    Parameters
    ----------
    d: float
        Particle position.
    l: float
        Box vector.

    Returns
    -------
    float
        Corrected particle position."""
    if np.abs(d) > 0.5 * l:
        d *= 1 - l / np.abs(d)
    return d

def calculate_temperature(number_of_particles, particles):
    """Determine the instantaneous temperature of the system.

    Parameters
    ----------
    system: System
        Whole system information.

    Returns
    -------
    Particle array:
        All particles updated with their velocities scaled.
    System:
        Whole system information with the temperature updated."""
    k = 0
    for i in range(0, number_of_particles):
        v = np.sqrt(particles['xvelocity'][i] * particles['xvelocity'][i] + particles['yvelocity'][i] * particles['yvelocity'][i])
        k += 0.5 * v * v
    return k / number_of_particles

def particle_dt():
    return np.dtype([('xposition', np.float64), ('yposition', np.float64), ('xvelocity', np.float64), ('yvelocity', np.float64), ('xacceleration', np.float64), ('yacceleration', np.float64), ('xprevious_position', np.float64), ('yprevious_position', np.float64), ('xforce', np.float64), ('yforce', np.float64), ('energy', np.float64)])
